# This file is part of ap_association.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import tempfile
import unittest
from unittest.mock import patch, Mock, MagicMock, DEFAULT
import warnings

import numpy as np
import pandas as pd

import lsst.afw.image as afwImage
import lsst.afw.table as afwTable
import lsst.dax.apdb as daxApdb
import lsst.pex.config as pexConfig
import lsst.utils.tests
from lsst.pipe.base.testUtils import assertValidOutput

from lsst.ap.association import DiaPipelineTask
from utils_tests import makeExposure, makeDiaObjects


def _makeMockDataFrame():
    """Create a new mock of a DataFrame.

    Returns
    -------
    mock : `unittest.mock.Mock`
        A mock guaranteed to accept all operations used by `pandas.DataFrame`.
    """
    with warnings.catch_warnings():
        # spec triggers deprecation warnings on DataFrame, but will
        # automatically adapt to any removals.
        warnings.simplefilter("ignore", category=DeprecationWarning)
        return MagicMock(spec=pd.DataFrame())


class TestDiaPipelineTask(unittest.TestCase):

    @classmethod
    def _makeDefaultConfig(cls,
                           config_file,
                           doPackageAlerts=False,
                           doSolarSystemAssociation=False):
        config = DiaPipelineTask.ConfigClass()
        config.doConfigureApdb = False
        config.apdb_config_url = config_file
        config.doPackageAlerts = doPackageAlerts
        config.doSolarSystemAssociation = doSolarSystemAssociation
        return config

    def setUp(self):
        # schemas are persisted in both Gen 2 and Gen 3 butler as prototypical catalogs
        srcSchema = afwTable.SourceTable.makeMinimalSchema()
        srcSchema.addField("base_PixelFlags_flag", type="Flag")
        srcSchema.addField("base_PixelFlags_flag_offimage", type="Flag")
        self.srcSchema = afwTable.SourceCatalog(srcSchema)

        apdb_config = daxApdb.ApdbSql.init_database(db_url="sqlite://")
        self.config_file = tempfile.NamedTemporaryFile()
        self.addCleanup(self.config_file.close)
        apdb_config.save(self.config_file.name)

    # TODO: remove on DM-43419
    def testConfigApdbNestedOk(self):
        config = DiaPipelineTask.ConfigClass()
        config.doConfigureApdb = True
        with self.assertWarns(FutureWarning):
            config.apdb.db_url = "sqlite://"
            config.freeze()
            config.validate()

    # TODO: remove on DM-43419
    def testConfigApdbNestedInvalid(self):
        config = DiaPipelineTask.ConfigClass()
        config.doConfigureApdb = True
        # Don't set db_url
        config.freeze()
        with self.assertRaises(pexConfig.FieldValidationError):
            config.validate()

    # TODO: remove on DM-43419
    def testConfigApdbFileOk(self):
        config = DiaPipelineTask.ConfigClass()
        config.doConfigureApdb = False
        config.apdb_config_url = "some/file/path.yaml"
        config.freeze()
        config.validate()

    # TODO: remove on DM-43419
    def testConfigApdbFileInvalid(self):
        config = DiaPipelineTask.ConfigClass()
        config.doConfigureApdb = False
        # Don't set apdb_config_url
        config.freeze()
        with self.assertRaises(pexConfig.FieldValidationError):
            config.validate()

    def testRun(self):
        """Test running while creating and packaging alerts.
        """
        self._testRun(doPackageAlerts=True, doSolarSystemAssociation=True)

    def testRunWithSolarSystemAssociation(self):
        """Test running while creating and packaging alerts.
        """
        self._testRun(doPackageAlerts=False, doSolarSystemAssociation=True)

    def testRunWithAlerts(self):
        """Test running while creating and packaging alerts.
        """
        self._testRun(doPackageAlerts=True, doSolarSystemAssociation=False)

    def testRunWithoutAlertsOrSolarSystem(self):
        """Test running without creating and packaging alerts.
        """
        self._testRun(doPackageAlerts=False, doSolarSystemAssociation=False)

    def _testRun(self, doPackageAlerts=False, doSolarSystemAssociation=False):
        """Test the normal workflow of each ap_pipe step.
        """
        config = self._makeDefaultConfig(
            config_file=self.config_file.name,
            doPackageAlerts=doPackageAlerts,
            doSolarSystemAssociation=doSolarSystemAssociation)
        task = DiaPipelineTask(config=config)
        # Set DataFrame index testing to always return False. Mocks return
        # true for this check otherwise.
        task.testDataFrameIndex = lambda x: False
        diffIm = Mock(spec=afwImage.ExposureF)
        exposure = Mock(spec=afwImage.ExposureF)
        template = Mock(spec=afwImage.ExposureF)
        diaSrc = _makeMockDataFrame()
        ssObjects = _makeMockDataFrame()
        ccdExposureIdBits = 32

        # Each of these subtasks should be called once during diaPipe
        # execution. We use mocks here to check they are being executed
        # appropriately.
        subtasksToMock = [
            "diaCatalogLoader",
            "diaCalculation",
            "diaForcedSource",
        ]
        if doPackageAlerts:
            subtasksToMock.append("alertPackager")
        else:
            self.assertFalse(hasattr(task, "alertPackager"))

        if not doSolarSystemAssociation:
            self.assertFalse(hasattr(task, "solarSystemAssociator"))

        def concatMock(_data, **_kwargs):
            return _makeMockDataFrame()

        # Mock out the run() methods of these two Tasks to ensure they
        # return data in the correct form.
        def solarSystemAssociator_run(unAssocDiaSources, solarSystemObjectTable, diffIm):
            return lsst.pipe.base.Struct(nTotalSsObjects=42,
                                         nAssociatedSsObjects=30,
                                         ssoAssocDiaSources=_makeMockDataFrame(),
                                         unAssocDiaSources=_makeMockDataFrame())

        def associator_run(table, diaObjects, exposure_time=None):
            return lsst.pipe.base.Struct(nUpdatedDiaObjects=2, nUnassociatedDiaObjects=3,
                                         matchedDiaSources=_makeMockDataFrame(),
                                         unAssocDiaSources=_makeMockDataFrame(),
                                         longTrailedSources=None)

        # apdb isn't a subtask, but still needs to be mocked out for correct
        # execution in the test environment.
        with patch.multiple(task, **{task: DEFAULT for task in subtasksToMock + ["apdb"]}), \
            patch('lsst.ap.association.diaPipe.pd.concat', side_effect=concatMock), \
            patch('lsst.ap.association.association.AssociationTask.run',
                  side_effect=associator_run) as mainRun, \
            patch('lsst.ap.association.ssoAssociation.SolarSystemAssociationTask.run',
                  side_effect=solarSystemAssociator_run) as ssRun:

            result = task.run(diaSrc,
                              ssObjects,
                              diffIm,
                              exposure,
                              template,
                              ccdExposureIdBits,
                              "g")
            for subtaskName in subtasksToMock:
                getattr(task, subtaskName).run.assert_called_once()
            assertValidOutput(task, result)
            # Exact type and contents of apdbMarker are undefined.
            self.assertIsInstance(result.apdbMarker, pexConfig.Config)
            meta = task.getFullMetadata()
            # Check that the expected metadata has been set.
            self.assertEqual(meta["diaPipe.numUpdatedDiaObjects"], 2)
            self.assertEqual(meta["diaPipe.numUnassociatedDiaObjects"], 3)
            # and that associators ran once or not at all.
            mainRun.assert_called_once()
            if doSolarSystemAssociation:
                ssRun.assert_called_once()
            else:
                ssRun.assert_not_called()

    def test_createDiaObjects(self):
        """Test that creating new DiaObjects works as expected.
        """
        nSources = 5
        diaSources = pd.DataFrame(data=[
            {"ra": 0.04*idx, "dec": 0.04*idx,
             "diaSourceId": idx + 1 + nSources, "diaObjectId": 0,
             "ssObjectId": 0}
            for idx in range(nSources)])

        config = self._makeDefaultConfig(config_file=self.config_file.name, doPackageAlerts=False)
        task = DiaPipelineTask(config=config)
        result = task.createNewDiaObjects(diaSources)
        self.assertEqual(nSources, len(result.newDiaObjects))
        self.assertTrue(np.all(np.equal(
            result.diaSources["diaObjectId"].to_numpy(),
            result.diaSources["diaSourceId"].to_numpy())))
        self.assertTrue(np.all(np.equal(
            result.newDiaObjects["diaObjectId"].to_numpy(),
            result.diaSources["diaSourceId"].to_numpy())))

    def test_purgeDiaObjects(self):
        """Remove diaOjects that are outside an image's bounding box.
        """

        config = self._makeDefaultConfig(config_file=self.config_file.name, doPackageAlerts=False)
        task = DiaPipelineTask(config=config)
        exposure = makeExposure(False, False)
        nObj0 = 20

        # Create diaObjects
        diaObjects = makeDiaObjects(nObj0, exposure)
        # Shrink the bounding box so that some of the diaObjects will be outside
        bbox = exposure.getBBox()
        size = np.minimum(bbox.getHeight(), bbox.getWidth())
        bbox.grow(-size//4)
        exposureCut = exposure[bbox]
        sizeCut = np.minimum(bbox.getHeight(), bbox.getWidth())
        buffer = 10
        bbox.grow(buffer)

        def check_diaObjects(bbox, wcs, diaObjects):
            raVals = diaObjects.ra.to_numpy()
            decVals = diaObjects.dec.to_numpy()
            xVals, yVals = wcs.skyToPixelArray(raVals, decVals, degrees=True)
            selector = bbox.contains(xVals, yVals)
            return selector

        selector0 = check_diaObjects(bbox, exposureCut.getWcs(), diaObjects)
        nIn0 = np.count_nonzero(selector0)
        nOut0 = np.count_nonzero(~selector0)
        self.assertEqual(nObj0, nIn0 + nOut0)

        diaObjects1 = task.purgeDiaObjects(exposureCut.getBBox(), exposureCut.getWcs(), diaObjects,
                                           buffer=buffer)
        # Verify that the bounding box was not changed
        sizeCheck = np.minimum(exposureCut.getBBox().getHeight(), exposureCut.getBBox().getWidth())
        self.assertEqual(sizeCut, sizeCheck)
        selector1 = check_diaObjects(bbox, exposureCut.getWcs(), diaObjects1)
        nIn1 = np.count_nonzero(selector1)
        nOut1 = np.count_nonzero(~selector1)
        nObj1 = len(diaObjects1)
        self.assertEqual(nObj1, nIn0)
        # Verify that not all diaObjects were removed
        self.assertGreater(nObj1, 0)
        # Check that some diaObjects were removed
        self.assertLess(nObj1, nObj0)
        # Verify that no objects outside the bounding box remain
        self.assertEqual(nOut1, 0)
        # Verify that no objects inside the bounding box were removed
        self.assertEqual(nIn1, nIn0)


class MemoryTester(lsst.utils.tests.MemoryTestCase):
    pass


def setup_module(module):
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()
